Option Strict On
Option Explicit On 

Imports System.Runtime.InteropServices
Imports System.Text

Friend Class DialogFolderSelect
    Implements IDisposable

    '========================================================================
    'Interfaces and Functions to communicate with the WINDOWS through shell32 
    '========================================================================

    Private Const MAX_PATH As Integer = 260

    <InterfaceType(ComInterfaceType.InterfaceIsIUnknown), _
     GuidAttribute("00000002-0000-0000-C000-000000000046")> _
    Private Interface IMalloc
        <PreserveSig()> Function Alloc(ByVal cb As Integer) As IntPtr
        <PreserveSig()> Function Realloc(ByVal pv As IntPtr, ByVal cb As Integer) As IntPtr
        <PreserveSig()> Sub Free(ByVal pv As IntPtr)
        <PreserveSig()> Function GetSize(ByVal pv As IntPtr) As Integer
        <PreserveSig()> Function DidAlloc(ByVal pv As IntPtr) As Integer
        <PreserveSig()> Sub HeapMinimize()
    End Interface

    'This function must be called before the SHBrowseForFolder function call
    'when the Bffstyles.NewDialogStyle option is specified.
    Private Declare Function OleInitialize Lib "ole32" _
                                  (ByVal pvReserved As IntPtr) As Integer

    Private Declare Function SHGetMalloc Lib "shell32" _
                                  (ByRef ppMalloc As IMalloc) As Integer

    Private Declare Function SHBrowseForFolder Lib "shell32" _
                                  (ByRef lpbi As BROWSEINFO) As IntPtr

    Private Declare Ansi Function SHGetPathFromIDList Lib "shell32" Alias "SHGetPathFromIDListA" _
                                  (ByVal pidl As IntPtr, ByVal pszPath As StringBuilder) As Integer

    Private Declare Function SHGetSpecialFolderLocation Lib "shell32" _
                                  (ByVal hWnd As IntPtr, ByVal nFolder As Integer, _
                                   ByRef ppidl As IntPtr) As Integer

    Private Declare Ansi Function SendMessage Lib "user32" Alias "SendMessageA" _
                                  (ByVal hWnd As IntPtr, ByVal Msg As Integer, _
                                   ByVal wParam As Integer, ByVal lParam As IntPtr) As Integer

    <StructLayout(LayoutKind.Sequential, Pack:=8)> _
    Private Structure BROWSEINFO
        Public hwndOwner As IntPtr
        Public pidlTopFolder As IntPtr
        Public pszDisplayName As IntPtr
        <MarshalAs(UnmanagedType.LPStr)> _
        Public lpszTitle As String
        Public ulFlags As Integer
        <MarshalAs(UnmanagedType.FunctionPtr)> _
        Public lpfn As CallbackProc
        Public lParam As IntPtr
        Public iImage As Integer
    End Structure

    Public Enum BffStyles
        RestrictToFileSystem = &H1      'For finding a folder to start document searching
        RestrictToDomain = &H2          'For starting the Find Computer
        ShowStatusText = &H4
        RestrictToSubFolders = &H8
        ShowTextBox = &H10
        ValidateSelection = &H20        'insist on valid result (or CANCEL)
        NewDialogStyle = &H40
        NewUserInterface = NewDialogStyle + ShowTextBox
        BrowseURLs = &H80
        ShowUsageHint = &H100
        NoNewFolderButton = &H200
        NoShortcutTranslate = &H400
        BrowseForComputer = &H1000      'Browsing for Computers.
        BrowseForPrinter = &H2000       'Browsing for Printers
        BrowseForEverything = &H4000    'Browsing for Everything
        BrowseShareableResources = &H8000
    End Enum

    Public Enum FolderID
        Desktop = &H0
        Internet = &H1
        Programs = &H2
        Controls = &H3
        Printers = &H4
        MyDocuments = &H5
        Favorites = &H6
        Startup = &H7
        Recent = &H8
        Sendto = &H9
        BitBucket = &HA
        StartMenu = &HB
        DesktopDirectory = &H10
        MyComputer = &H11
        NetworkNeighborhood = &H12
        Nethood = &H13
        Fonts = &H14
        Templates = &H15
        CommonStartMenu = &H16
        CommonPrograms = &H17
        CommonStartup = &H18
        CommonDesktopDirectory = &H19
        AppDdata = &H1A
        Printhood = &H1B
        AltStartup = &H1D               'DBCS
        CommonAltStartup = &H1E         'DBCS
        CommonFavorites = &H1F
        InternetCache = &H20
        Cookies = &H21
        History = &H22
        MyPictures = &H27
        NetAndDialUpConnections = &H31
    End Enum

    Private Delegate Function CallbackProc _
                             (ByVal hwnd As IntPtr, ByVal uMsg As Integer, _
                              ByVal lParam As IntPtr, ByVal lpData As IntPtr) As Integer

    'message from browser
    Private Const BFFM_INITIALIZED As Integer = 1
    Private Const BFFM_SELCHANGED As Integer = 2
    Private Const BFFM_VALIDATEFAILEDA As Integer = 3   'lParam:szPath ret:1(cont),0(EndDialog)
    Private Const BFFM_VALIDATEFAILEDW As Integer = 4   'lParam:wzPath ret:1(cont),0(EndDialog)

    'messages to browser
    Private Const WM_USER As Integer = &H400
    Private Const BFFM_SETSTATUSTEXTA As Integer = (WM_USER + 100)
    Private Const BFFM_ENABLEOK As Integer = (WM_USER + 101)
    Private Const BFFM_SETSELECTIONA As Integer = (WM_USER + 102)
    Private Const BFFM_SETSELECTIONW As Integer = (WM_USER + 103)
    Private Const BFFM_SETSTATUSTEXTW As Integer = (WM_USER + 104)

    '======================================
    'Private properties of this class's own
    '======================================
    Private Const ME_NAME As String = "DialogFolderSelect"

    Private Disposed As Boolean = False

    Private mstrStartFolder As String = Nothing
    Private mpidlStartFolder As IntPtr = IntPtr.Zero

    Private mstrDescription As String = Nothing

    Private msfTopFolder As FolderID = FolderID.Desktop
    Private mbfsStyle As BffStyles = BffStyles.RestrictToFileSystem

    Private mstrReturnPath As String = Nothing
    Private mstrReturnDisplayName As String = Nothing

    Private mfrmOwner As Form

    Public Sub New(ByVal frmOwner As Form)
        mfrmOwner = frmOwner
    End Sub

    Public Sub SetStartFolder(ByVal Value As String)
        If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
        If Not mpidlStartFolder.Equals(IntPtr.Zero) Then
            Marshal.FreeCoTaskMem(mpidlStartFolder)
            mpidlStartFolder = IntPtr.Zero
        End If
        mstrStartFolder = Value
    End Sub

    Public Sub SetStartFolder(ByVal Value As FolderID)
        If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
        mstrStartFolder = Nothing
        If Not mpidlStartFolder.Equals(IntPtr.Zero) Then
            Marshal.FreeCoTaskMem(mpidlStartFolder)
            mpidlStartFolder = IntPtr.Zero
        End If

        Dim rc As Integer = SHGetSpecialFolderLocation(mfrmOwner.Handle, Value, mpidlStartFolder)
        If rc <> 0 Then Marshal.ThrowExceptionForHR(rc)
    End Sub

    Public Property TopFolder() As FolderID
        Get
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            Return msfTopFolder
        End Get
        Set(ByVal Value As FolderID)
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            msfTopFolder = Value
        End Set
    End Property

    Public Property Style() As BffStyles
        Get
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            Return mbfsStyle
        End Get
        Set(ByVal Value As BffStyles)
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            mbfsStyle = Value
        End Set
    End Property

    Public Property Description() As String
        Get
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            Return mstrDescription
        End Get
        Set(ByVal Value As String)
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            mstrDescription = Value
        End Set
    End Property

    Public Function ShowDialog() As DialogResult
        Dim bi As BROWSEINFO
        Dim buffer As IntPtr = Marshal.AllocHGlobal(MAX_PATH)
        Dim rc As Integer
        Dim bytDebug As Byte

        If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)

        Dim pidlTopFolder As IntPtr
        rc = SHGetSpecialFolderLocation(mfrmOwner.Handle, msfTopFolder, pidlTopFolder)
        If rc <> 0 Then Marshal.ThrowExceptionForHR(rc)

        With bi
            .hwndOwner = mfrmOwner.Handle
            .pidlTopFolder = pidlTopFolder
            .pszDisplayName = buffer
            .lpszTitle = mstrDescription
            .ulFlags = mbfsStyle
            .lpfn = New CallbackProc(AddressOf BrowseFolderCallback)
        End With

        If (bi.ulFlags And BffStyles.NewDialogStyle) > 0 Then OleInitialize(IntPtr.Zero)

        Dim pidl As IntPtr = SHBrowseForFolder(bi)

        If pidl.Equals(IntPtr.Zero) Then
            mstrReturnDisplayName = ""
            mstrReturnPath = ""
            ShowDialog = DialogResult.Cancel
        Else
            Dim pszPath As StringBuilder = New StringBuilder(MAX_PATH)
            rc = SHGetPathFromIDList(pidl, pszPath)
            mstrReturnPath = pszPath.ToString

            mstrReturnDisplayName = Marshal.PtrToStringAnsi(bi.pszDisplayName)

            ShowDialog = DialogResult.OK
        End If

        Marshal.FreeHGlobal(buffer)

        Dim malloc As IMalloc
        SHGetMalloc(malloc)

#If Debug Then
        Debug.WriteLine("pidl size:" & malloc.GetSize(pidl).ToString)
        If Not pidl.Equals(IntPtr.Zero) Then DumpIDL(pidl)
#End If

        malloc.Free(pidl)
        malloc.Free(pidlTopFolder)
    End Function

    Private Function BrowseFolderCallback _
                    (ByVal hWnd As IntPtr, ByVal uMsg As Integer, _
                     ByVal lParam As IntPtr, ByVal lpData As IntPtr) As Integer

        Dim rc As Integer

        Select Case uMsg
            Case BFFM_INITIALIZED
                If Not mstrStartFolder Is Nothing Then
                    Dim pszPath As IntPtr = Marshal.StringToHGlobalAnsi(mstrStartFolder)
                    rc = SendMessage(hWnd, BFFM_SETSELECTIONA, -1&, pszPath)
                    Marshal.FreeHGlobal(pszPath)
                ElseIf Not mpidlStartFolder.Equals(IntPtr.Zero) Then
                    rc = SendMessage(hWnd, BFFM_SETSELECTIONA, 0&, mpidlStartFolder)
                End If
            Case BFFM_SELCHANGED
                Dim sbPath As StringBuilder = New StringBuilder(MAX_PATH)
                rc = SHGetPathFromIDList(lParam, sbPath)

                Dim pszPath As IntPtr = Marshal.StringToHGlobalAnsi(sbPath.ToString)
                rc = SendMessage(hWnd, BFFM_SETSTATUSTEXTA, 0, pszPath)
                Marshal.FreeHGlobal(pszPath)
        End Select
    End Function

    Public ReadOnly Property Path() As String
        Get
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            Return mstrReturnPath
        End Get
    End Property

    Public ReadOnly Property DisplayName() As String
        Get
            If Me.Disposed Then Throw New ObjectDisposedException(ME_NAME)
            Return mstrReturnDisplayName
        End Get
    End Property

#Region "for Debugging"
#If Debug Then
    Private Sub DumpIDL(ByVal ptr As IntPtr)
        Dim ofs As Integer = 0
        Dim TotalLength As Integer = 0

        Dim cb1, cb2 As Byte

        Debug.WriteLine("--------------------------------------------")
        Do
            cb1 = Marshal.ReadByte(ptr, ofs)
            ofs += 1
            cb2 = Marshal.ReadByte(ptr, ofs)
            ofs += 1

            If cb1 = 0 AndAlso cb2 = 0 Then Exit Do
            Debug.Write("Size : ")
            Debug.WriteLine(cb1 + cb2 * 256)
            TotalLength += cb1 + cb2 * 256

            Dim I As Integer
            For I = 1 To cb1 + cb2 * 256 - 2
                Debug.Write((I + 2).ToString.PadLeft(4) & ":")
                Debug.WriteLine(Marshal.ReadByte(ptr, ofs).ToString.PadLeft(4))
                ofs += 1
            Next
            Debug.WriteLine("---------")
        Loop
        Debug.WriteLine("Total Length : " & TotalLength.ToString)
    End Sub
#End If
#End Region

    Public Overloads Sub Dispose() Implements IDisposable.Dispose
        Dispose(True)
        GC.SuppressFinalize(Me)
    End Sub

    Protected Overridable Overloads Sub Dispose(ByVal disposing As Boolean)
        If Not Me.Disposed Then
            If disposing Then
                mfrmOwner = Nothing
            Else
                If Not mpidlStartFolder.Equals(IntPtr.Zero) Then
                    Dim malloc As IMalloc
                    SHGetMalloc(malloc)
                    malloc.Free(mpidlStartFolder)
                    mpidlStartFolder = IntPtr.Zero
                End If
            End If
        End If
        Me.Disposed = True
    End Sub

    Protected Overrides Sub Finalize()
        Dispose(False)
    End Sub
End Class
