Option Strict On
Option Explicit On 

Imports System.Net
Imports System.IO
Imports System.Text
Imports System.Text.RegularExpressions
Imports vb = Microsoft.VisualBasic

Public Class HttpConnection
    Const BUFFER_SIZE As Integer = 1024

    Private mParent As Control = Nothing

    Public Event HeaderReceived(ByVal sender As Object, ByVal e As HttpConEventArgs)
    Public Event DataReceiving(ByVal sender As Object, ByVal e As HttpConEventArgs)
    Public Event DataReceived(ByVal sender As Object, ByVal e As HttpConEventArgs)
    Public Event ErrorEncountered(ByVal sender As Object, ByVal e As HttpConErrorEventArgs)

    Private Delegate Sub DelegateForInvoke(ByVal sender As Object, ByVal e As HttpConEventArgs)
    Private Delegate Sub DelegateForInvokeError(ByVal sender As Object, ByVal e As HttpConErrorEventArgs)

    Protected Sub OnHeaderReceived(ByVal sender As Object, ByVal e As HttpConEventArgs)
        RaiseEvent HeaderReceived(sender, e)
    End Sub

    Protected Sub OnDataReceiving(ByVal sender As Object, ByVal e As HttpConEventArgs)
        RaiseEvent DataReceiving(sender, e)
    End Sub

    Protected Sub OnDataReceived(ByVal sender As Object, ByVal e As HttpConEventArgs)
        RaiseEvent DataReceived(sender, e)
    End Sub

    Protected Sub OnErrorEncountered(ByVal sender As Object, ByVal e As HttpConErrorEventArgs)
        RaiseEvent ErrorEncountered(sender, e)
    End Sub

    Public Sub New(ByVal Parent As Form)
        mParent = Parent
    End Sub

    Public Sub SendRequestGET(ByVal hwrRequest As HttpWebRequest)
        hwrRequest.Method = "GET"

        hwrRequest.BeginGetResponse(New AsyncCallback(AddressOf RespCallback), _
                                    New RequestState(hwrRequest))
    End Sub

    Public Sub SendRequestPOST(ByVal hwrRequest As HttpWebRequest, ByVal bytPostData() As Byte)
        hwrRequest.Method = "POST"
        hwrRequest.ContentLength = bytPostData.Length

        hwrRequest.BeginGetRequestStream(New AsyncCallback(AddressOf ReqStreamCallback), _
                                         New RequestState(hwrRequest, bytPostData))
    End Sub

    Private Sub ReqStreamCallback(ByVal ar As IAsyncResult)
        Dim rs As RequestState = CType(ar.AsyncState, RequestState)

        Dim srRequestData As Stream
        Try
            srRequestData = rs.Request.EndGetRequestStream(ar)
        Catch ex As Exception
            Dim DelegateErrorArgs() As Object = {Me, New HttpConErrorEventArgs(rs.Request, ex)}
            If mParent Is Nothing Then
                OnErrorEncountered(DelegateErrorArgs(0), _
                                   CType(DelegateErrorArgs(1), HttpConErrorEventArgs))
            Else
                mParent.Invoke(New DelegateForInvokeError(AddressOf Me.OnErrorEncountered), _
                               DelegateErrorArgs)
            End If
            Exit Sub
        End Try

        srRequestData.Write(rs.PostData, 0, rs.PostData.Length)
        srRequestData.Close()

        ' Issue the async request.
        rs.Request.BeginGetResponse(New AsyncCallback(AddressOf RespCallback), rs)
    End Sub

    Private Sub RespCallback(ByVal ar As IAsyncResult)
        Dim rs As RequestState = CType(ar.AsyncState, RequestState)

        Dim resp As HttpWebResponse
        Try
            resp = CType(rs.Request.EndGetResponse(ar), HttpWebResponse)
        Catch we As WebException
            If we.Status = WebExceptionStatus.ProtocolError Then
                resp = CType(we.Response, HttpWebResponse)
            Else
                resp.Close()

                Dim DelegateErrorArgs() As Object = {Me, New HttpConErrorEventArgs(rs.Request, we)}
                If mParent Is Nothing Then
                    OnErrorEncountered(DelegateErrorArgs(0), _
                                       CType(DelegateErrorArgs(1), HttpConErrorEventArgs))
                Else
                    mParent.Invoke(New DelegateForInvokeError(AddressOf Me.OnErrorEncountered), _
                                   DelegateErrorArgs)
                End If
                Exit Sub
            End If
        End Try

        rs.Response = resp
        rs.BufferStream = New MemoryStream()

        Dim DelegateArgs() As Object = {Me, New HttpConEventArgs(rs.Request, resp, Nothing)}
        If mParent Is Nothing Then
            OnHeaderReceived(DelegateArgs(0), CType(DelegateArgs(1), HttpConEventArgs))
        Else
            mParent.Invoke(New DelegateForInvoke(AddressOf Me.OnHeaderReceived), DelegateArgs)
        End If

        If CType(DelegateArgs(1), HttpConEventArgs).Cancel Then Exit Sub

        rs.ResponseStream = resp.GetResponseStream()
        rs.ResponseStream.BeginRead(rs.BufferRead, 0, BUFFER_SIZE, _
                                    New AsyncCallback(AddressOf ReadCallBack), rs)
    End Sub

    Private Sub ReadCallBack(ByVal asyncResult As IAsyncResult)
        Dim rs As RequestState = CType(asyncResult.AsyncState, RequestState)
        Dim intLength As Integer = rs.ResponseStream.EndRead(asyncResult)

        If intLength > 0 Then
            rs.BufferStream.Write(rs.BufferRead, 0, intLength)

            Dim DelegateArgs() As Object = {Me, New HttpConEventArgs(rs.Request, rs.Response, rs.BufferStream)}
            If mParent Is Nothing Then
                OnDataReceiving(DelegateArgs(0), CType(DelegateArgs(1), HttpConEventArgs))
            Else
                mParent.Invoke(New DelegateForInvoke(AddressOf Me.OnDataReceiving), DelegateArgs)
            End If

            If CType(DelegateArgs(1), HttpConEventArgs).Cancel Then
                rs.ResponseStream.Close()
                rs.Response.Close()
                Exit Sub
            End If

            rs.ResponseStream.BeginRead(rs.BufferRead, 0, BUFFER_SIZE, _
                                        New AsyncCallback(AddressOf ReadCallBack), rs)
        Else
            Dim DelegateArgs() As Object = {Me, New HttpConEventArgs(rs.Request, rs.Response, rs.BufferStream)}
            If mParent Is Nothing Then
                OnDataReceived(DelegateArgs(0), CType(DelegateArgs(1), HttpConEventArgs))
            Else
                mParent.Invoke(New DelegateForInvoke(AddressOf Me.OnDataReceived), DelegateArgs)
            End If
        End If
    End Sub

    Private Class RequestState
        Public Request As HttpWebRequest = Nothing

        Public Response As HttpWebResponse = Nothing
        Public ResponseStream As Stream = Nothing
        Public BufferRead(BUFFER_SIZE) As Byte

        Public PostData As Byte() = Nothing

        Public BufferStream As MemoryStream = Nothing

        Public Sub New(ByVal value As HttpWebRequest)
            Request = value
        End Sub

        Public Sub New(ByVal value As HttpWebRequest, ByVal Data() As Byte)
            Request = value
            PostData = Data
        End Sub

        Public Sub New()

        End Sub
    End Class
End Class

Public Class HttpConEventArgs
    Inherits EventArgs

    Private mhwrRequest As HttpWebRequest
    Private mhwrResponse As HttpWebResponse
    Private msResponseStream As MemoryStream
    Private mblnCancel As Boolean

    Public Sub New(ByVal Request As HttpWebRequest, _
                   ByVal Response As HttpWebResponse, _
                   ByVal ResponseStream As MemoryStream)
        MyBase.New()
        mhwrRequest = Request
        mhwrResponse = Response
        msResponseStream = ResponseStream
    End Sub

    Public ReadOnly Property Request() As HttpWebRequest
        Get
            Return mhwrRequest
        End Get
    End Property

    Public ReadOnly Property Response() As HttpWebResponse
        Get
            Return mhwrResponse
        End Get
    End Property

    Public ReadOnly Property ResponseStream() As MemoryStream
        Get
            Return msResponseStream
        End Get
    End Property

    Public Property Cancel() As Boolean
        Get
            Return mblnCancel
        End Get
        Set(ByVal Value As Boolean)
            mblnCancel = Value
        End Set
    End Property

End Class

Public Class HttpConErrorEventArgs
    Inherits EventArgs

    Private mhwrRequest As HttpWebRequest
    Private meException As exception

    Public Sub New(ByVal Request As HttpWebRequest, _
                   ByVal e As exception)
        MyBase.New()
        mhwrRequest = Request
        meException = e
    End Sub

    Public ReadOnly Property Request() As HttpWebRequest
        Get
            Return mhwrRequest
        End Get
    End Property

    Public ReadOnly Property exception() As exception
        Get
            Return meException
        End Get
    End Property
End Class
